#!/usr/bin/lua

RECEIVE_BLOCK_SIZE = 2^10
MAX_RECEIVE_BLOCKS = 2
WINDOW_SIZE = 8

local uitcpConfDefault = {
  tunnelListener = {
    host = "*",
    port = 0,
  },
  tunnel = {
    connectAttempt = 24,
    connectTmo = 5,
    keepalive = 3,
    attempt = 3,
    window = 8,
  },
}

local function x(s)
  os.execute("("..s..")>/dev/null 2>&1")
end

local ipAddrTable = {}
local function ipAddrAdd(addr)
  local n = ipAddrTable[addr]
  if n then
    ipAddrTable[addr] = n + 1
  else
    x("ip addr add "..addr.."/32 dev eth0")
    ipAddrTable[addr] = 1
  end
end
local function ipAddrDel(addr)
  local n = ipAddrTable[addr]
  if n then
    if n <= 1 then
      ipAddrTable[addr] = nil
    else
      ipAddrTable[addr] = n - 1
      return
    end
  end
  x("ip addr del "..addr.."/32 dev eth0")
end

-- Get user parameters
--pcall(loadfile"/etc/uitcp.conf")
local configFile = "/etc/uitcp.conf"
local f, e = loadfile(configFile)
if not f then
  print("Error loading config file", configFile)
  error(e)
end
local res, err = pcall(f)
if not res then
  print("Error in config file", configFile)
  error(err)
end



local conf = uitcpConf or {}
setmetatable(conf, uitcpConfDefault)
uitcpConfDefault.__index = uitcpConfDefault
if conf.tunnel then
  setmetatable(conf.tunnel, uitcpConfDefault.tunnel)
  uitcpConfDefault.tunnel.__index = uitcpConfDefault.tunnel
end
--------------------------------------------------------------------------------
require "classlib"
--require"printtable"
local socket = require("socket")
require"ssl"
local dbg = conf.dbg
local log = conf.log
local trace = conf.trace
local logFile
if type(conf.logFile) == "string" then
  logFile = io.open(conf.logFile, "a")
else
  logFile = io.output()
end
if not logFile then log, trace = nil, nil end
function DBG(...) if dbg then print(...) end end
function LOG(...) if log then --print("LOG",...)
                     logFile:write(string.format("%s: %s\n",
                         os.date(), string.format(...)))
                     logFile:flush()
                     end end
function hex_dump(buf, ...)
  if trace then
    if type(trace) ~= "number" then trace = #buf end
    n = math.min(#buf, trace)
    logFile:write(string.format("%s: %s(%d bytes)\n",
                  os.date(), string.format(...), #buf))
    for byte=1, n, 16 do
       local chunk = buf:sub(byte, byte+15)
       logFile:write(string.format('%08X  ',byte-1))
       chunk:gsub('.', function (c) logFile:write(string.format('%02X ',string.byte(c))) end)
       logFile:write(string.rep(' ',3*(16-#chunk)))
       logFile:write(' ',chunk:gsub('[%c\255]','.'),"\n") 
    end
    logFile:flush()
  end
end

--------------------------------------------------------------------------------
hdr = {
  lenLow = 1,
  lenHigh = 2,
  dstLC = 3,
  srcLC = 4,
  sn = 5,
  rn = 6,
  len = 6,
}

threads = {}
sleepThr = {}
tunnels = {}
local createConnection
local openTcpSocket
local createRecvDataThread
local connectionCreated
local startSslHandshake
local createLocalListener
--------------------------------------------------------------------------------
local function readpopen(prog)
  local f = io.popen(prog)
  local s = f:read("*a")
  f:close()
  return s
end

local function isInterfaceUp(int)
  return string.find(readpopen("ip link show "..int), "UP")
end

--------------------------------------------------------------------------------
-- init timer
local timerList = {}

local function addTimer(tmo, func, ...)
  tmo = os.time() + tmo
  local timer = {time = tmo, func = func, arg = {...}}
  for i, t in ipairs(timerList) do
    if t.time > tmo then
      table.insert(timerList, i, timer)
      return timer
    end
  end
  table.insert(timerList, timer)
  return timer
end

local function delTimer(timer)
  if timer then
    for i, t in ipairs(timerList) do
      if t == timer then table.remove(timerList, i) end
    end
  end
end

local function getTimer()
  return timerList[1]
end

local function popTimer()
  return table.remove(timerList, 1)
end

local function execTimer(timer)
  timer.func(unpack(timer.arg))
end

local function timerHandler()
  local t = os.time()
  while true do
    local timer = getTimer()
    if not timer then return end
    if timer.time <= t then
      execTimer(popTimer())
    else return end
  end
end

MAX_TMO = 5
local function nextTmo()
  local t = os.time()
  local timer = getTimer()
  if not timer or timer.time - t > MAX_TMO then return MAX_TMO end
  return timer.time - t
end
    
--------------------------------------------------------------------------------
local function stopRecv(th)
  DBG("stopRecv", th)
  if th then sleepThr[th] = true end
end

local function startRecv(th)
  DBG("startRecv", th)
  sleepThr[th] = nil
  if coroutine.status(th) ~= "dead" then
    for i, t in ipairs(threads) do
      if t == th then
        DBG"allready exist"
        return
      end    
    end
    table.insert(threads, th)
    end
end

--------------------------------------------------------------------------------
class.Fifo()

function Fifo:__init()
  self.first = 0
  self.last = -1
end

function Fifo:push(v)
  local last = self.last + 1
  self.last = last
  self[last] = v
end

function Fifo:pop()
  local first = self.first
  if first > self.last then -- FIFO is empty
    return nil
  end
  local v = self[first]
  self[first] = nil
  self.first = first + 1
  return v
end
 
function Fifo:isEmpty()
  if self.first > self.last then
    return true
  else return false end
end
    
--------------------------------------------------------------------------------
class.Connection()

function Connection:__init(initial)
  DBG"Connection __init"
  self.sock = initial.sock
  self.server = initial.server
end

function Connection:__call(recvData)
  DBG("receive data", self.sock, recvData)
end

function Connection:getSock()
  return self.sock
end

function Connection:setSock(s)
  self.sock = s
end

function Connection:isServer()
  --DBG("isServer()", self.server)
  return self.server
end

function Connection:setThread(co)
  self.thread = co
end

function Connection:getThread(co)
  return self.thread
end

--------------------------------------------------------------------------------
class.TunnelConnection(Connection)

function TunnelConnection:__init(initial)
  self.Connection:__init(initial)
  DBG"TunnelConnection __init"
  self.state = "init"
  self.name = initial.name
  self.host = initial.host
  self.port = initial.port
  self.window = initial.window or WINDOW_SIZE
  self.keepTmo = initial.keepTmo
  self.keepAttempt = initial.keepAttempt
  self.ssl = initial.ssl
  self.nat = initial.nat
  self:init()
  if not self:isServer() then
    tunnels[self.name] = self
    self.links = initial.links
    self.linkN = #self.links
    self.linkLast = self.linkN
    self.linkCurrent = self.linkN
    self.linksPriority = initial.linksPriority
    self:connect()
  end
end

function TunnelConnection:init()
  self.state = "init"
  self.s = {} -- table for slave connection
  self.listenerSocks = {}
  self.sended = {}
  self.pending = Fifo()
  self.recvBuf = ""
  self.sn = 0
  self.rn = 0
  self.an = 0
  self.rr = 0
  self.connectErr = 0
  self.lastRecvTime = os.time()
end

function TunnelConnection:sendCommandNew()
  self:send('m:new("'..self.name..'",'..self.window..','..
            self.keepTmo..','..self.keepAttempt..','..
            self.keepTmo * (self.keepAttempt+1) + 
            conf.tunnel.connectTmo * (conf.tunnel.connectAttempt+1)..')')
end

function TunnelConnection:reinit()
  self:allSlave('destroy')
  self:init()
  if not self:isServer() then
    self:sendCommandNew()
    self:initSession()
  end
end

function TunnelConnection:initSession()
  self.state = "waitcon"
  self.keepSeq = 0
  self.keepErr = 0
  self.connectErr = 0
end

function TunnelConnection:new(name, w, tmo, attempt, watchdog)
  if self.state == "init" or self.state == "disc" then
    if tunnels[name] then
      LOG("Tunnel "..name.." already exist. Destroy it")
      tunnels[name]:destroy()
    end
    tunnels[name] = self
    self.name = name
    self.nat = conf.clients[name].nat
    self.ssl = conf.clients[name].ssl
    createLocalListener(self, conf.clients[name].localListener)
    self.window = w
    self.keepTmo = tmo
    self.keepAttempt = attempt
    self.watchdogTmo = watchdog
    self:send('m:newack("'..self.name..'")')
    LOG("New tunnel %s initiated from %s", name, self:getSock():getpeername())
    self:opened()
  else
    error("Get new() in not init or disc state")
  end
end

function TunnelConnection:newack(id)
  if self.state == "waitcon" then
    LOG("New tunnel %s initiated", self.name)
    self:opened()
  else
    error("get newack() in not waitcon state")
  end
end

function TunnelConnection.connected(c, argt)
  DBG("Tunnel Connected")
  local self = argt.self
  if c then
    self.linkCurrent = self.linkN
    if self.linksPriority then self.linkN = #self.links end
    self.linkLast = self.linkN
    self:setSock(c)
    createRecvDataThread(self)
    if self.state == "init" then
      self:sendCommandNew()
      LOG("Init new tunnel %s", self.name)
    else
      self:send('m:resume("'..self.name..'")')
      LOG("Resume tunnel %s", self.name)
    end
    self:initSession()
  else
    LOG("Can't connect host %s port %s gw %s", self.host, self.port,
      self.links[self.linkN].dev)
    if self.linkLast == self.linkN then
      self.connectErr = self.connectErr + 1
      DBG("connectErr", self.connectErr)
      if self.connectErr >= conf.tunnel.connectAttempt then
        LOG("Can't restore tunnel \"%s\", all local connection will be closed",
            self.name)
        self:allSlave('destroy')
        self:init()
      end
      self.connectTimer = addTimer(conf.tunnel.connectTmo, self.connect, self)
    else
      self:connect()
    end
  end
 
end

function TunnelConnection:connect()
  DBG("Tunnel Connect", self.host, self.port, #self.links, self.linkN)
  local link = self.links[self.linkN]
  if link and link.restartScript then
    if not link.lastScriptMark then
       link.lastScriptMark = os.time()
       if type(link.restartScript) == "string" then
         link.restartScript = {link.restartScript}
       end
       if type(link.restartScript) ~= "table" then
         link.restartScript = nil
       else
         link.scriptCurr = 0
         link.scriptNum = #link.restartScript
       end
    else  
      if not link.restartScriptTmo or
         os.time() - link.lastScriptMark >= link.restartScriptTmo then
        link.scriptCurr =   link.scriptCurr + 1
        if link.scriptCurr > link.scriptNum then link.scriptCurr = 1 end
        LOG("Start script: %s", link.restartScript[link.scriptCurr])   
        x(link.restartScript[link.scriptCurr])
        link.lastScriptMark = os.time()
      end
    end
  end
  self.linkN = self.linkN + 1
  if self.linkN > #self.links then self.linkN = 1 end
  LOG("Try connect host %s port %s gw %s", self.host, self.port,
      self.links[self.linkN].dev)
  x('ip route del default; ip route add default dev '..
             self.links[self.linkN].dev)
  local c = openTcpSocket(self.host, self.port,
                          TunnelConnection.connected, {self = self})
end

function TunnelConnection:resume(name)
  if tunnels[name] then
    tunnels[name]:resumeConnection(self:getSock(), self.an)
    self.dead = true
  else
    self:send('m:resnack("'..name..'")')
    LOG("Can't resume tunnel %s, no such tunnel", name)
  end
end

function TunnelConnection:resumeConnection(sock, an)
  self.an = an
  self:closed()
  LOG("Tunnel \"%s\" resumed from %s", self.name, sock:getpeername())
  self:setSock(sock)
  self:send('m:resack("'..self.name..'")')
  createRecvDataThread(self)
  self:opened()
  --self:rej(self.an)
end  

function TunnelConnection:resack(name)
  DBG("resack", name)
  if self.state == "waitcon" then
    LOG("Tunnel %s resumed", name)
    self:opened()
    --self:rej(self.an)
  else
    error("get resack() in not waitcon state")
  end
end

function TunnelConnection:resnack(name)
  LOG("Tunnel %s not resumed, try reinit tunnel", name)
  self:reinit()
end

function TunnelConnection:__call(recvData)
  hex_dump(recvData, "receive data from tunnel %s", self.name or "<NONE>")
  self.recvBuf = self.recvBuf..recvData
  self.lastRecvTime = os.time()
  while true do
    local buf = self.recvBuf
    if #buf < hdr.len then break end
    local len = buf:byte(hdr.lenLow) + buf:byte(hdr.lenHigh) * 256
    if len > #buf - hdr.len then break end -- too many data, wait

    self.recvBuf = self.recvBuf:sub(len + hdr.len + 1) -- remove head
    local dst, src = buf:byte(hdr.dstLC), buf:byte(hdr.srcLC)
    local data = buf:sub(hdr.len + 1, len + hdr.len)
    self.an = buf:byte(hdr.rn)
--[[
    if dst == 0 and src == 0 then -- command
      DBG("tunCom: ", data)
      if len > 0 then
        local f, e = loadstring("local m=...;"..data)
        if not f then DBG(e) error(e) end
        res, err = pcall(f, self)
        if not res then error(err) end
      end
    else -- data
      local bufSn = buf:byte(hdr.sn)
      if bufSn ~= self.rn then 
        self.recvBuf = ""
        self:send('m:rej('..self.rn..')')
      else 
        self.rn = (self.rn + 1) % 256
        local slave = self.s[dst]
        if slave then
          slave:send(data)
        else
          LOG("No such LC: %d", dst)
        end
      end
    end
--]]
-- [[
    local rej = false
    if not (dst == 0 and src == 0) then -- numbered packet
      local bufSn = buf:byte(hdr.sn)
      if bufSn ~= self.rn then 
        self.recvBuf = ""
        self:send('m:rej('..self.rn..')')
        rej = true
      else 
        self.rn = (self.rn + 1) % 256
      end
    end
    if not rej then
      if dst == 0 then -- command
        DBG("tunCom: ", data)
        if len > 0 then
          local f, e = loadstring("local m=...;"..data)
          if not f then DBG(e) error(e) end
          res, err = pcall(f, self)
          if not res then error(err) end
        end
      else -- data
        local slave = self.s[dst]
        if slave then
          slave:send(data)
        else
          LOG("No such LC: %d", dst)
        end
      end
    end
--]]
    
  end -- while
  self:send() -- may be window has opened
end

function TunnelConnection:realSend(data, remoteLC, localLC, sn)
  local sock = self:getSock()
  if not sock then return nil end
  local len = #data
  local hd = string.char(len % 256, math.floor(len / 256),
                         remoteLC, localLC, sn, self.rn)
  self.rr = self.rn
  local packet =  hd..data                    
  hex_dump(packet, "Send to tunnel %s", self.name or "<NONE>")
  local n, e, s, x = sock:send(packet)
  if not n then
    DBG("SEND ERROR", n, e, s, x)
    --error"SEND ERROR"
  end
  return n
end

function TunnelConnection:sendRR()
  if (self.rn - self.rr) % 256 >= self.window then
    self:realSend("", 0, 0, self.sn)
  end
end

function TunnelConnection:send(data, remoteLC, localLC)
  DBG("send to tun", self.name, self.sn, self.an, remoteLC, localLC)
  if data and not remoteLC then -- send unnumbered packet
    return self:realSend(data, 0, 0, self.sn)
  else -- send numbered packet
    if (self.sn - self.an) % 256 >= self.window or -- no window
       self.state ~= "data" then
      DBG("no window")
      if data then
        --if self.pending:isEmpty() then self:allSlave('sleep') end
        self:allSlave('sleep')
        self.pending:push{data, remoteLC, localLC}
      end
      self:sendRR()
    elseif data and not self.pending:isEmpty() then
      DBG"pending"
      self.pending:push{data, remoteLC, localLC}
      self:send()
    else
      DBG"data"
      if not self.pending:isEmpty() then
        data, remoteLC, localLC = unpack(self.pending:pop())
        if self.pending:isEmpty() then self:allSlave('wake') end
      end
      if data then
        self:realSend(data, remoteLC, localLC, self.sn) -- TODO(if return nil)
        self.sended[self.sn] = {data, remoteLC, localLC, self.sn}
        self.sn = (self.sn + 1) % 256
        self:send()
      end
      self:sendRR()
    end
  end
  return true
end

function TunnelConnection:rej(n)
  -- TODO if sended[n] already nil
  DBG("rej", n, self.sn)
  while n ~= self.sn do
    DBG(n)
    self:realSend(unpack(self.sended[n])) -- TODO -- bad argument #1 to 'unpack' (table expected, got nil)
    n = (n + 1) % 256
  end
  self:send()
end

function TunnelConnection:addSlave(s)
  DBG("addSlave", self, s)
  local lc = #self.s + 1
  self.s[lc] = s -- TODO find free LC
  s.localLC = lc
  if s:isServer() then
    local srcHost, srcPort = s:getSock():getpeername()
    local dstHost, dstPort = s:getSock():getsockname()
    LOG("Accept local connection(src = %s:%d, dst = %s:%d), lc = %d",
        srcHost, srcPort, dstHost, dstPort, lc)
    local com = 'm:con('..lc..',"'..srcHost..'",'..dstPort..')'
    if not self:send(com, 0, lc) then
      s:destroy()
      self.s[lc] = nil
    end
  end
end

function TunnelConnection.conCreated(c, conf)
  DBG("conCreated", c)
  local s = connectionCreated(c, conf)
  local lc = 0
  if s then
    lc = s.localLC
  else
    LOG("Can't connect %s:%d", conf.host, conf.port)
  end
  local com = 'm:conack('..conf.remoteLC..','..lc..')'
  conf.master:send(com, 0, 255)
end

function TunnelConnection:con(remoteLC, inSrcAddr, inDstPort)
  DBG("con:", remoteLC, inSrcAddr, inDstPort)
  LOG("Local connection request from \"%s\", inSrcAddr = %s, inDstPort = %d",
      self.name, inSrcAddr, inDstPort)
  local host
  local port
  local srcHost
  if type(conf.localCon) == "table" then
    host = conf.localCon.outDstAddr
    port = conf.localCon.outDstPort
    srcHost = conf.localCon.outSrcAddr
  end
  if self.nat then
    for _, v in ipairs(self.nat) do
      DBG(v.inSrcAddr, inDstPort, v.outDstAddr, v.outDstPort, v.outSrcAddr)
      if (not v.inSrcAddr or v.inSrcAddr == inSrcAddr) and
         (not v.inDstPort or v.inDstPort == inDstPort) then
        host = v.outDstAddr or host
        port = v.outDstPort or port
        srcHost = v.outSrcAddr or srcHost
        break
      end
    end
  end
  if not host or not port then
    LOG("Can't find parameters for setting local connection")
    local com = 'm:conack('..remoteLC..','..'0'..')'
    self:send(com, 0, 255)
    return
  end
  LOG("Set local connection to %s:%d, src = %s", host, port, srcHost or "")
  local arg = {
    host = host,
    port = port,
    srcHost = srcHost,
    remoteLC = remoteLC,
    conClass = LocalConnection,
    master = self,
    remoteLC = remoteLC,
    func = TunnelConnection.conCreated,
  }
  createConnection(arg)
end

function TunnelConnection:conack(dst, src)
  if not self.s[dst] then
    LOG("Connect ack: No such local connection, lc = %d", dst)
    return
  end
  if src == 0 then
    LOG("Not connected lc = %d", dst)
    self.s[dst]:destroy()
    self.s[dst] = nil
  else
    LOG("Connected lc = %d. remote lc = %d", dst, src)
    self.s[dst]:setRemoteLC(src)
  end
end

function TunnelConnection:closeReq(s)
  local num = false
  for k, v in pairs(self.s) do -- TODO ipairs
    if v == s then num = k; break end
  end
  if num then 
    LOG("Local connection %d closed", s.localLC)
    local com = 'm:close('..s.remoteLC..')'
    self:send(com, 0, s.localLC)
  end
end

function TunnelConnection:close(n)
  LOG("Local connection %d closed by request from tunnel %s", n, self.name)
  if self.s[n] then
    local com = 'm:closeack('..self.s[n].remoteLC..')'
    self:send(com, 0, self.s[n].localLC)
    self.s[n]:destroy()
    self.s[n] = nil
  end
end

function TunnelConnection:closeack(n)
  self.s[n] = nil
end

function TunnelConnection:tryReOpen()
  if self.linkCurrent > 1 and self.keepSeq > 10 then
    return true
  else
    return false
  end
end

function TunnelConnection:keepalive()
  if os.time() - self.lastRecvTime >= self.keepTmo then
    self.keepSeq = self.keepSeq + 1
    self.keepErr = self.keepErr + 1
    if  self.keepErr > self.keepAttempt or
        not isInterfaceUp(self.links[self.linkCurrent].dev) or
        self:tryReOpen() then 
      LOG("Peer not answered")    
      --self:closed()
      self.noRecv = true
      self:getSock():close()
      return
    end
    local com = 'm:keep('..self.keepSeq..')'
    self:send(com)
  end
  self.keepTimer = addTimer(self.keepTmo, self.keepalive, self)
end

function TunnelConnection:watchdog()
  local silence = os.time() - self.lastRecvTime
  if silence >= self.watchdogTmo then
    LOG("Can't restore tunnel \"%s\", all local connection will be closed",
        self.name)
    self.noRecv = true
    self:destroy()
  else
    self.keepTimer = addTimer(self.watchdogTmo - silence, self.watchdog, self)
  end
end

function TunnelConnection:keep(n)
  local com = 'm:keepack('..tostring(n)..')'
  self:send(com)
end

function TunnelConnection:keepack(n)
  if n == self.keepSeq then self.keepErr = 0 end
end

function TunnelConnection:closed()
  LOG("Tunnel \"%s\" closed", self.name or "<NONE>")
  self.noRecv = nil
  self.state = "disc"
  delTimer(self.keepTimer)
  self:getSock():close()
  self.recvBuf = ""
  self:allSlave('sleep')
  if not self:isServer() then
    self:connect()
  end
end

function TunnelConnection:startDataPhase()
  self.state = "data"
  self:allSlave('wake')
  self:rej(self.an)
  if not self:isServer() then
    self.keepTimer = addTimer(self.keepTmo, self.keepalive, self)
  else
    self.keepTimer = addTimer(self.watchdogTmo or 180, self.watchdog, self)
  end
end

function TunnelConnection.doneSslHandshake(self, res, err)
  local self = self
  delTimer(self.sslDoHandshakeTimer)
  if res then
    LOG("SSL handshake successfully done")
    createRecvDataThread(self)
    self:startDataPhase()
  else
    LOG("Can't do ssl handshake: %s", err)
    self:getSock():close()
    self:closed()
  end
end

function TunnelConnection:opened()
  DBG("Tunnel opened", self)
  if self.ssl then
    LOG("Start SSL handshake over tunnel \"%s\"", self.name)
    local res, err
    local sock = self:getSock()
    DBG("sock", sock)
    res, sock, err = pcall(ssl.wrap, sock, self.ssl)
    if res and sock then
      self:setSock(sock)
      DBG("SSL sock", sock, self.keepAttempt * self.keepTmo)
      startSslHandshake(self, TunnelConnection.doneSslHandshake)
      self.sslDoHandshakeTimer = addTimer(self.keepAttempt * self.keepTmo,
                                          self.doneSslHandshake,
                                          self, nil, "Timeout exhausted")
    else
      LOG("Can't do ssl wrap: %s", sock or err)
      self:getSock():close()
    end
    return
  end
  
  self:startDataPhase()
end

function TunnelConnection:destroy()
  self.dead = true
  self:getSock():close()
  delTimer(self.keepTimer)
  self:allSlave('destroy')
  for _, sock in ipairs(self.listenerSocks) do
    sock:close()
  end
  tunnels[self.name] = nil
end

function TunnelConnection:allSlave(func, ...)
  for k, v in ipairs(self.s) do
    v[func](v, ...)
  end
end
  
--------------------------------------------------------------------------------
class.LocalConnection(Connection)

function LocalConnection:__init(initial)
  DBG"LocalConnection __init"
  self.Connection:__init(initial)
  self.localLC = 0
  self.remoteLC = initial.remoteLC or 0
  self.master = initial.master
  self.srcHost = initial.srcHost
  self.pending = Fifo()
  self.master:addSlave(self)
end

function LocalConnection:__call(recvData)
  if not recvData then
    if not self.pending:isEmpty() and
       self.remoteLC ~= 0 and  self.localLC ~= 0 then
      self.master:send(self.pending:pop(), self.remoteLC, self.localLC)
      self()
    end
  else -- recvData
   if dbg then hex_dump(recvData, "receive data from local") end
   if self.remoteLC ~= 0 and self.localLC ~= 0 and self.pending:isEmpty() then
      self.master:send(recvData, self.remoteLC, self.localLC)
    else
      self.pending:push(recvData)
      self()
    end
  end
end

function LocalConnection:send(data)
  if dbg then hex_dump(data, "send to local") end
  repeat
    local len, s, n, e = #data, 0
    n, e, s = self:getSock():send(data, s + 1)
    if not n then
      DBG("send error", e, s)
      if e ~= "timeout" then return end -- TODO
    end
  until n or s >= len
end

function LocalConnection:setRemoteLC(lc)
  self.remoteLC = lc
end

function LocalConnection:destroy()
  DBG("destroy", self)
  self:getSock():close()
  if self.srcHost then
    ipAddrDel(self.srcHost)
  end
end

function LocalConnection:closed()
  DBG("closed", self)
  self:destroy()
  self.master:closeReq(self)
end

function LocalConnection:sleep()
  stopRecv(self:getThread())
end

function LocalConnection:wake()
  startRecv(self:getThread())
end
--------------------------------------------------------------------------------

class.ControlConnection(Connection)

function ControlConnection:__init(initial)
  DBG"LocalConnection __init"
  self.Connection:__init(initial)
  self.srcHost = initial.srcHost
end

function ControlConnection:__call(recvData)
   if dbg then hex_dump(recvData, "receive data from control") end
end

function ControlConnection:send(data)
  if dbg then hex_dump(data, "send to control") end
  repeat
    local len, s, n, e = #data, 0
    n, e, s = self:getSock():send(data, s + 1)
    if not n then
      DBG("send error", e, s)
      if e ~= "timeout" then return end -- TODO
    end
  until n or s >= len
end

function ControlConnection:closed()
  DBG("ControlConnection:closed", self)
end

---------------Listen stuff-----------------------------------------------------
local function accept(s)
  s:settimeout(0)
  local c, status = s:accept()
  if status == "timeout" then
    coroutine.yield(s)
  end
  return c, status
end

local function listen(conf)
  while true do
    local c, status = accept(conf.serverSock)
    if c then
      local con = conf.conClass{sock = c, master = conf.master, server = conf.server}
      DBG("New connection", con, c)
      createRecvDataThread(con)
    elseif status == "closed" then
      DBG("listener closed", conf.host, conf.port)
      break
    end
  end
end

local function createListenThread(conf)
  local s = assert(socket.bind(conf.host, conf.port))
  local i, p = s:getsockname()
  assert(i, p)
  LOG("Waiting connection from talker on %s:%d", i, p)
  conf.serverSock = s
  local co = coroutine.create(function()
                                listen(conf)
                              end)
  table.insert(threads, co)
  return s
end
-----------------Receive stuf---------------------------------------------------
local function recvData(con, sock)
  DBG("recvData, con =", con)
  DBG("sock =", sock)
  local n = 0
  while true do
    sock:settimeout(0)
    local data, status, partial = sock:receive(RECEIVE_BLOCK_SIZE)
    if sock ~= con:getSock() then
      sock:close()
      DBG("recvData closed", con)
      return
    end
    data = data or partial
    if #data > 0 then
      DBG("recvData", con)
      if not con.noRecv then con(data) end
      if con.dead then DBG"dead" return end -- when resume tunnel connection
      n = n + 1
    end
    if status == "timeout" or
       status == "wantread" or
       status == "wantwrite" or
       n >= MAX_RECEIVE_BLOCKS then
      coroutine.yield(sock, status == "wantwrite")
      n = 0
    elseif status == "closed" then break end
  end
  DBG("closed")
  con:closed()
end

function createRecvDataThread(con)
  local sock = con:getSock()
  DBG("createRecvDataThread", con, sock)
  local co = coroutine.create(function()
                                recvData(con, sock)
                              end)
  table.insert(threads, co)
  con:setThread(co)
end

--------------------------------------------------------------------------------
local function waitTcpSocket(host, port, func, argt)
  local m = socket.tcp()
  if argt.srcHost then
    ipAddrAdd(argt.srcHost)
    if not m:bind(argt.srcHost, 0) then
      ipAddrDel(argt.srcHost)
      m:close()
      func(nil, argt)
      return
    end
  end
  while true do
    m:settimeout(0)
    local c, status = m:connect(host, port)
    DBG("waitTcpSocket", c, status)
    if c then
      func(m, argt)
      return
    elseif status == "timeout" --[[or
           status == "Operation already in progress"]] then -- TODO (bad criterium)
      coroutine.yield(m, true)
    else
      m:close()
      func(nil, argt)
      return
    end
  end
end

function openTcpSocket(host, port, func, argt)
  local co = coroutine.create(function()
                                waitTcpSocket(host, port, func, argt)
                              end)
  table.insert(threads, co)
end

function connectionCreated(m, conf)
  if not m then return nil end
  conf.sock = m
  local con = conf.conClass(conf)
  createRecvDataThread(con)
  return con
end

function createConnection(conf)
  DBG"createConnection"
  openTcpSocket(conf.host, conf.port, conf.func, conf)
end

--------------------------------------------------------------------------------

local function waitSslHandShake(self, func, sock)
  while true do
    sock:settimeout(0)
    local res, err = sock:dohandshake()
    DBG("waitSslHandShake", res, err)
    if sock ~= self:getSock() then
      sock:close()
      DBG("waitSslHandShake closed(no sock)", con)
      return
    end
    if res then
      func(self, res, err)
      return
    elseif err == "wantread" or err == "wantwrite" or err == "timeout" then
      coroutine.yield(sock,  status == "wantwrite")
    else
      sock:close()
      func(self, res, err)
      return
    end
  end
end

function startSslHandshake(self, func)
  local sock = self:getSock()
  local co = coroutine.create(function()
                                waitSslHandShake(self, func, sock)
                              end)
  table.insert(threads, co)
end
--------------------------------------------------------------------------------
local function dispatch()
  local i = 1
  local connections = {}
  local sndConnections = {}
  local status, res, snd
  while true do
    if threads[i] == nil then
      if threads[1] == nil then break end
      i = 1
      connections = {}
      sndConnections = {}
    end
    if sleepThr[threads[i]] then
      res = nil
    else
      --print("dispatch resume", threads[i])
      status, res, snd = coroutine.resume(threads[i])
      --print("dispatch", status, res, snd)
    end
    if not status then
      error("coroutine failed\n"..res)
    end
    if not res then
      table.remove(threads, i)
    else -- timeout
      i = i + 1
      if snd then
        sndConnections[#sndConnections + 1] = res
      else
        connections[#connections + 1] = res
      end
      if #connections + #sndConnections == #threads then
        DBG("start select", #threads)
          local recvt, sendt, err = socket.select(connections, sndConnections, nextTmo())
          if gbg then print(os.time()) end
          timerHandler()
          if dbg then
            for _, v in ipairs(recvt) do print(v) end
            for _, v in ipairs(sendt) do print(v) end
          end
      end
    end
  end -- end while
end

--------------------------------------------------------------------------------
function createLocalListener(master, localListenerTable)
  if not master or not localListenerTable then return end
  for _, v in ipairs(localListenerTable) do
    local arg = {
      host = v.host,
      port = v.port,
      conClass = LocalConnection,
      master = master,
      server = true,
    }
    table.insert(master.listenerSocks, createListenThread(arg))
  end
end

local function createTunnelListener()
  local arg = {
    host = conf.tunnelListener.host,
    port = conf.tunnelListener.port,
    conClass = TunnelConnection,
    server = true,
  }
  createListenThread(arg)
end

local function createControlListener()
  local arg = {
    host = "*",  --conf.tunnelListener.host,
    port = 8888, --conf.tunnelListener.port,
    conClass = ControlConnection,
  }
  createListenThread(arg)
end

local function createTunnel()
  local arg = {
    name = conf.tunnel.name,
    host = conf.tunnel.host,
    port = conf.tunnel.port,
    window = conf.tunnel.window,
    keepTmo = conf.tunnel.keepalive,
    keepAttempt = conf.tunnel.attempt,
    links = conf.tunnel.links,
    linksPriority = conf.tunnel.linksPriority,
    ssl = conf.tunnel.ssl,
    nat = conf.tunnel.nat,
  }
  return TunnelConnection(arg)
end

--------------------------------------------------------------------------------
createControlListener()
if conf.createTunnel then
  mainTunnel = createTunnel()
  createLocalListener(mainTunnel, conf.tunnel.localListener)
else
  createTunnelListener()
end
dispatch()
--------------------------------------------------------------------------------
--[[
iptables -A OUTPUT -t nat -p tcp -d 192.168.10.1 --dport 8080 -j DNAT --to-destination 82.138.15.92
--]]
